// Netify Agent Log Sink Plugin
// Copyright (C) 2021-2023 eGloo Incorporated
// <http://www.egloo.ca>
//
// This program is free software: you can redistribute it
// and/or modify it under the terms of the GNU General
// Public License as published by the Free Software
// Foundation, either version 3 of the License, or (at your
// option) any later version.
//
// This program is distributed in the hope that it will be
// useful, but WITHOUT ANY WARRANTY; without even the
// implied warranty of MERCHANTABILITY or FITNESS FOR A
// PARTICULAR PURPOSE.  See the GNU General Public License
// for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program.  If not, see
// <http://www.gnu.org/licenses/>.

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <fstream>

#include "nsp-plugin.hpp"

void nspChannelConfig::Load(const string &channel,
                            const json &jconf) {
  auto it = jconf.find("overwrite");
  if (it != jconf.end() &&
      it->type() == json::value_t::boolean)
    overwrite = it->get<bool>();

  it = jconf.find("log_path");
  if (it != jconf.end() &&
      it->type() == json::value_t::string)
    log_path = it->get<string>();

  it = jconf.find("log_name");
  if (it != jconf.end() &&
      it->type() == json::value_t::string)
    log_name = it->get<string>();
}

nspPlugin::nspPlugin(const string &tag,
                     const ndPlugin::Params &params)
    : ndPluginSink(tag, params) {
  reload = true;

  nd_dprintf("%s: initialized\n", tag.c_str());
}

nspPlugin::~nspPlugin() {
  Join();

  nd_dprintf("%s: destroyed\n", tag.c_str());
}

void *nspPlugin::Entry(void) {
  nd_printf(
      "%s: %s v%s (C) 2021-2023 eGloo Incorporated.\n",
      tag.c_str(), PACKAGE_NAME, PACKAGE_VERSION);

  while (!ShouldTerminate()) {
    if (reload.load()) {
      Reload();
      reload = false;
    }

    if (WaitOnPayloadQueue()) {
      ndPluginSinkPayload *p;
      while ((p = PopPayloadQueue())) {
#if 0
                nd_dprintf("%s: payload, length: %lu, %p\n",
                    tag.c_str(), p->length, p->data
                );
                for (auto &c : p->channels) {
                    nd_dprintf("%s: -> channel: %s\n",
                        tag.c_str(), c.c_str()
                    );
                }
#endif
        WritePayload(p);

        delete p;
      }
    }
  }

  return nullptr;
}

void nspPlugin::DispatchEvent(ndPlugin::Event event,
                              void *param) {
  switch (event) {
    case ndPlugin::EVENT_RELOAD:
      reload = true;
      break;
    default:
      break;
  }
}

void nspPlugin::Reload(void) {
  nd_dprintf("%s: Loading configuration: %s\n", tag.c_str(),
             conf_filename.c_str());

  json j;

  ifstream ifs(conf_filename);
  if (!ifs.is_open()) {
    nd_printf("%s: Error loading configuration: %s: %s\n",
              tag.c_str(), conf_filename.c_str(),
              strerror(ENOENT));
    return;
  }

  try {
    ifs >> j;
  } catch (exception &e) {
    nd_printf("%s: Error loading configuration: %s: %s\n",
              "parse error", tag.c_str(),
              conf_filename.c_str());
    nd_dprintf("%s: %s: %s\n", tag.c_str(),
               conf_filename.c_str(), e.what());
    return;
  }

  Lock();

  defaults.Load("defaults", j);

  channels.clear();

  try {
    auto jchannels = j.find("channels");
    if (jchannels != j.end()) {
      for (auto &kvp : jchannels->get<json::object_t>()) {
        auto it = kvp.second.find("enable");
        if (it != kvp.second.end()) {
          if (it->type() == json::value_t::boolean &&
              it->get<bool>() != true)
            break;
        }

        nspChannelConfig config;
        config.Load(kvp.first, kvp.second);

        if (config.log_name.empty()) {
          throw ndPluginException("log_name",
                                  strerror(EINVAL));
        }

        channels.insert(make_pair(kvp.first, config));
      }
    }
  } catch (exception &e) {
    Unlock();
    throw e;
  }

  Unlock();
}

void nspPlugin::WritePayload(ndPluginSinkPayload *payload) {
  for (auto &channel : payload->channels) {
    auto channel_conf = channels.find(channel);
    if (channel_conf == channels.end()) {
      nd_printf("%s: channel not found: %s\n", tag.c_str(),
                channel.c_str());
      continue;
    }

    string log_suffix(".log");
    if ((payload->flags & ndPlugin::DF_FORMAT_JSON))
      log_suffix = ".json";
    else if ((payload->flags & ndPlugin::DF_FORMAT_MSGPACK))
      log_suffix = ".msgpack";

    string ext;
    if ((payload->flags & ndPlugin::DF_GZ_DEFLATE))
      ext = ".gz";

    ndLogDirectory ld(channel_conf->second.log_path,
                      channel_conf->second.log_name,
                      log_suffix,
                      channel_conf->second.overwrite);

    FILE *hf;
    if ((hf = ld.Open(ext))) {
      fwrite(payload->data, 1, payload->length, hf);

      ld.Close();
    } else {
      nd_dprintf("%s: %s: Error opening new log file: %s\n",
                 tag.c_str(), channel.c_str(),
                 strerror(errno));
    }
  }
}

ndPluginInit(nspPlugin);
